#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <dirent.h>
#include <fcntl.h>
#include <inttypes.h>
#include <glob.h>

#include <sys/mman.h>
#include <sys/types.h>

#include <xenctrl.h>
#include <xen/elfnote.h>
#include <xen/elfstructs.h>

#include "xenner.h"

/* -------------------------------------------------------------- */

static int do_stats(char *filename)
{
    struct stat        st;
    int                fd;
    
    void               *vmcore;
    off_t              vmcore_size;
    Elf64_Ehdr         *vm_ehdr;
    Elf64_Phdr         *vm_phdrs;
    Elf64_Shdr         *vm_shdrs;
    char               *vm_shstrtab;
    
    uint64_t           last_faults[XEN_FAULT_MAX];
    uint64_t           last_hcalls[XEN_HCALL_MAX];

    uint64_t *faults, *hcalls;
    char *name;
    void *ptr;
    int i, rc = -1;
    
    fd = open(filename, O_RDWR);
    if (-1 == fd) {
	fprintf(stderr, "open %s: %s\n", filename, strerror(errno));
	goto cleanup;
    }

    fstat(fd, &st);
    vmcore_size = st.st_size;
    vmcore = mmap(NULL, VMCORE_HEADER + 1024 * PAGE_SIZE, PROT_READ | PROT_WRITE,
		  MAP_SHARED, fd, 0);
    if ((void*)-1 == vmcore) {
	fprintf(stderr, "%s: mmap(%s,%zd): %s\n",
		__FUNCTION__, filename, vmcore_size, strerror(errno));
	vmcore = NULL;
	goto cleanup;
    }

    vm_ehdr  = vmcore;
    vm_phdrs = vmcore + vm_ehdr->e_phoff;
    vm_shdrs = vmcore + vm_ehdr->e_shoff;
    vm_shstrtab = vmcore +
	vm_shdrs[vm_ehdr->e_shstrndx].sh_offset;

    for (i = 0; i < vm_ehdr->e_shnum; i++) {
	name = vm_shstrtab + vm_shdrs[i].sh_name;
	ptr  = vmcore + vm_shdrs[i].sh_offset;
	fprintf(stderr, "  %d: %s at %lx +%lx (%p)\n", i, name,
		vm_shdrs[i].sh_offset, vm_shdrs[i].sh_size, ptr);
	if (0 == strcmp(name, ".addr.emu_faults"))
	    faults = ptr;
	if (0 == strcmp(name, ".addr.emu_hcalls"))
	    hcalls = ptr;
    }

    if (!faults || !hcalls) {
	fprintf(stderr, "failed to parse vmcore file\n");
	goto cleanup;
    }

    rc = 0;
    fprintf(stderr, "<%s>\n", filename);
    print_hcall_stats(NULL, hcalls, NULL);
    print_fault_stats(NULL, faults, NULL);
    for (;;) {
	memcpy(last_hcalls, hcalls, sizeof(last_hcalls));
	memcpy(last_faults, faults, sizeof(last_faults));
	sleep(1);
	fprintf(stderr, "\n");
	if (-1 == stat(filename, &st)) {
	    fprintf(stderr, "vm gone\n");
	    break;
	}
	fprintf(stderr,
		"\x1b[H" /* home */
		"\x1b[J" /* cls  */
		"<%s>\n", filename);
	print_hcall_stats(NULL, hcalls, last_hcalls);
	print_fault_stats(NULL, faults, last_faults);
    }

cleanup:
    if (-1 != fd)
	close(fd);
    fd = -1;
    if (vmcore)
	munmap(vmcore, VMCORE_HEADER + 1024 * PAGE_SIZE);
    vmcore = NULL;
    return rc;
}

int main(int argc, char *argv[])
{
    glob_t g;

    if (argc == 2)
	return do_stats(argv[1]);

    for (;;) {
	if (0 == glob("/var/run/xenner/vmcore.*", 0, NULL, &g)) {
	    do_stats(g.gl_pathv[0]);
	    globfree(&g);
	} else {
	    fprintf(stderr, "no vm running\n");
	}
	sleep(1);
    }
}
